
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
 
#include "globdef.h"
#include "uidef.h"
#include "screendef.h"
#include "hwaredef.h"
#include "seldef.h"
#include "rusage.h"
#include "thrdef.h"
#include "fft1def.h"
#include "keyboard_def.h"

void show_hardware_gain(void);

#define MAX_EXTIO_FILES 16
#define BUFFER_SIZE 256
char extio_name[BUFFER_SIZE], extio_model[BUFFER_SIZE];

#define EXTIO_DLL_OPEN 1
#define EXTIO_CALLBACK_RUNNING 2

#if(OSNUM == OSNUM_WINDOWS)
#include <windows.h>
typedef int bool;
typedef bool (WINAPI *pOpenHW)(void);
typedef int (WINAPI *pSetHWLO)(long freq);
typedef void (WINAPI *pCloseHW)(void);
typedef long (WINAPI *pGetHWLO)(void);
typedef void (WINAPI *pSetCallback)(void (* Callback)(int,int,float,short*));
typedef long (WINAPI *pTuneChanged)(long);
typedef void (WINAPI *pSetRFGain)(int);
typedef int (WINAPI *pGetRFGain)(void);
typedef bool (WINAPI *pInitHW)(char *name, char *model,int *type);
typedef void (WINAPI *pShowGUI)(void);
typedef void (WINAPI *pHideGUI)(void);
typedef int (WINAPI *pStartHW)(long freq);
typedef void (WINAPI *pStopHW)(void);
typedef long (WINAPI *pGetHWSR)(void);
#endif

#if(OSNUM == OSNUM_LINUX)
#include <pthread.h> 
typedef int bool;
typedef bool (*pOpenHW)(void);
typedef int (*pSetHWLO)(long freq);
typedef void (*pCloseHW)(void);
typedef long (*pGetHWLO)(void);
typedef void (*pSetCallback)(void (* Callback)(int,int,float,short*));
typedef long (*pTuneChanged)(long);
typedef void (*pSetRFGain)(int);
typedef int (*pGetRFGain)(void);
typedef bool (*pInitHW)(char *name, char *model,int *type);
typedef void (*pShowGUI)(void);
typedef void (*pHideGUI)(void);
typedef int (*pStartHW)(long freq);
typedef void (*pStopHW)(void);
typedef long (*pGetHWSR)(void);
#endif



pOpenHW OpenHW;
pSetHWLO SetHWLO;
pCloseHW CloseHW;
pGetHWLO GetHWLO;
pSetCallback SetCallback;
pTuneChanged TuneChanged;
pSetRFGain SetRFGain;
pGetRFGain GetRFGain;
pInitHW InitHW;
pShowGUI ShowGUI;
pHideGUI HideGUI;
pStartHW StartHW;
pStopHW StopHW;
pGetHWSR GetHWSR;

#define EXTIO_FREQ_CHANGED  101
#define EXTIO_RF_GAIN_CHANGED  210
#define EXTIO_SAMPRATE_CHANGED 100


#define RF_GAIN_UP_LIMIT 44 //dB
#define RF_GAIN_LOW_LIMIT -6 //dB

#define EXTIO_INVERTED 1
#define EXTIO_NON_INVERTED 2
#define EXTIO_CALLBACK_ENABLE 4
#define EXTIO_SELECTION_VERIFIED 8



#define EXTIO_LIB_FILENAME "par_extio_libname" 

void write_extio_parfile(char *fnam)
{
FILE *file;
file=fopen(EXTIO_LIB_FILENAME,"w");
if(file == NULL)
  {
  lirerr(1236);
  return;
  }
clear_screen();
fprintf(file,"%s\n",fnam);
fclose(file);
lir_text(0,0,fnam);
lir_text(0,1,"Inverted spectrum? (Y/N):");
invspk:;
to_upper_await_keyboard();
if(lir_inkey == 'Y')
  {
  ui.use_extio=EXTIO_INVERTED;
  }
else
  {
  if(lir_inkey != 'N')goto invspk;
  ui.use_extio=EXTIO_NON_INVERTED;
  }
lir_text(0,1,"Enable the ExtIO callback for soundcard devices? (Y/N):");
callb:;
to_upper_await_keyboard();
if(lir_inkey == 'Y')
  {
  ui.use_extio|=EXTIO_CALLBACK_ENABLE;
  }
else
  {
  if(lir_inkey != 'N')goto callb;
  }
#define EXTIO_INVERTED 1
#define EXTIO_NON_INVERTED 2
#define EXTIO_CALLBACK_ENABLE 4
#define EXTIO_SELECTION_VERIFIED 8
lir_text(0,1,
  "The ExtIO of your choice is now stored in the file par_extio_libname.");
lir_text(0,2,
  "When you now press any key Linrad will exit.");
lir_text(0,4,  
  "The library of your choice will be loaded next time Linrad starts."); 
lir_text(0,6,
  "In case something is wrong with the Extio library it may be");
lir_text(0,7,   
  "necessary to remove the par_extio file in order to start Linrad again.");
lir_text(5,9,press_any_key);
await_processed_keyboard();
lirerr(0);
}

void get_library_filename(char *s)
{
FILE *file;
int i, k;
file=fopen(EXTIO_LIB_FILENAME,"r");
if(file == NULL)
  {
  ui.use_extio=0;
  sprintf(s,"Unknown file in %s",EXTIO_LIB_FILENAME);
  return;
  }
k=fread(s,1,BUFFER_SIZE,file);
if(k==BUFFER_SIZE)lirerr(1317);
fclose(file);
for(i=0; i<BUFFER_SIZE; i++)
  {
  if( s[i]==0 ||
      s[i]==10 ||
      s[i]==13)
    {
    s[i]=0;
    return;
    }    
  }
}

void get_extio_name(char *name)
{
int i,j;
char s[BUFFER_SIZE];
get_library_filename(&s[0]);
i=0;
while(s[i] != 0)
  {
  if(strncmp("ExtIO",&s[i],5) == 0)
    {
    j=0;
    while(s[i]!='.' && s[i]!=0 && j<40)
      {
      name[j]=s[i];
      i++;
      j++;
      }
    name[j]=0;  
    return;
    }
  i++;
  }
name[0]=0;
}      
    

void Callback_Winrad(int cnt, int status, float IQoffs, short *IQdata)
{
char *ch, *dh;
int *iz;
float *z;
int j;
long frequency = 0;
double fFreq;
(void) IQoffs;
if(kill_all_flag)return;
switch(status)
  {
  case EXTIO_FREQ_CHANGED:
  if(GetHWLO)
    {
    frequency = (*GetHWLO)();
    fFreq = (double) frequency / 1e6;
    fg.passband_center = fFreq;
    if(fg.passband_center == 0)fg.passband_center=7;
    set_hardware_rx_frequency();
    }
  break;

  case EXTIO_RF_GAIN_CHANGED:
  if(GetRFGain)
    {
    fg.gain = GetRFGain();
    if(fg.gain > RF_GAIN_UP_LIMIT) fg.gain = RF_GAIN_UP_LIMIT;
    if(fg.gain < RF_GAIN_LOW_LIMIT) fg.gain = RF_GAIN_LOW_LIMIT;
    fg.gain_increment = 1;
    show_hardware_gain();
    }
  break;
  
  case EXTIO_SAMPRATE_CHANGED:
  if(extio_type != 4)
    {
    if(!GetHWSR)
      {
      lirerr(131600);
      }
    if(ui.rx_ad_speed!=GetHWSR())extio_speed_changed=TRUE;
    }
  break;
  }
if(extio_type == 4)return;
if(cnt > 0 && thread_status_flag[THREAD_EXTIO_INPUT] == THRFLAG_ACTIVE)
  {
  switch(extio_type)
    {
    case 1:
    case 3:
    case 6:
    memcpy(&timf1_char[timf1p_sdr],IQdata,snd[RXAD].block_bytes);
    break;

    case 5:
    case 8:
    ch=(char*)IQdata;
    dh=&timf1_char[timf1p_sdr];
    j=(snd[RXAD].block_bytes>>2);
    while(j > 0)
      {
      j--;
      dh[4*j+3]=ch[3*j+2];
      dh[4*j+2]=ch[3*j+1];
      dh[4*j+1]=ch[3*j  ];
      dh[4*j  ]=0;
      } 
    break;

    case 7:
    iz=(int*)IQdata;
    z=(float*)&timf1_char[timf1p_sdr];
    j=(snd[RXAD].block_bytes>>2);
    while(j > 0)
      {
      j--;
      iz[j]=(int)(0x7fffffff*z[j]);
      } 
    break;
    }
  timf1p_sdr=(timf1p_sdr+snd[RXAD].block_bytes)&timf1_bytemask;
  lir_set_event(EVENT_EXTIO_RXREADY);
  }
}

#if(OSNUM == OSNUM_LINUX)
#include <dirent.h>
#include <sys/types.h>
#include <sys/param.h>
#include <dlfcn.h>
#include <semaphore.h>

#include "thrdef.h"
#include "lconf.h"
#include "ldef.h"

#if(IA64 == 0)
char *dirs[]={"./",
              "/lib",
              "/lib32",
              "/usr/lib",
              "/usr/lib32",
              "/usr/lib32/lib",
              "/usr/local/lib",
              "/usr/local32/lib",
              "/emul/ia32-linux/lib",
              "/emul/ia32-linux/usr/lib",              
              "/usr/lib/i386-linux-gnu",
              "X"};
#else
char *dirs[]={"./",
              "/lib",
              "/lib64",
              "/usr/lib",
              "/usr/lib64",
              "/usr/local/lib",
              "/usr/lib/x86_64-linux-gnu",
              "X"};
#endif



void init_extio(void)
{
char extio_lib_names[MAX_EXTIO_FILES][BUFFER_SIZE];
DIR *dirp;
char s[BUFFER_SIZE];
int i, m, num_extio;
int lineno;
struct dirent *dp;
i=0;  
lineno=0;
lir_text(10,1,"Seaching libraries for ExtIO files");
lir_refresh_screen();
while(dirs[i][0] != 'X' && lineno < MAX_EXTIO_FILES)
  {
  dirp=opendir(dirs[i]);
  if(dirp != NULL)
    {
next_file:;
    dp=readdir(dirp);
    if(dp != NULL)
      {
      m=strncmp("libExtIO",dp->d_name,8);
      if(m == 0)
        {
        sprintf(extio_lib_names[lineno],"%s/%s", dirs[i], dp->d_name);
        sprintf(s,"%2d %s",lineno,extio_lib_names[lineno]);
        lir_text(3,lineno+4,s);
        lineno++;
        }
      goto next_file;
      }
    closedir(dirp);  
    }
  i++;  
  }
if(lineno == 0)
  {
  lir_text(8,lineno+6,"No libExtIO_xxx.so file found");
  lir_text(5,lineno+8,press_any_key);
  await_keyboard();
  return;
  }  
else
  {
  lir_text(8,lineno+6,"Select libExtIO_xxx.so file by line number:");
  num_extio=lir_get_integer(51,lineno+6,2,0,lineno-1);  
  if(kill_all_flag)return;
  }
write_extio_parfile(extio_lib_names[num_extio]);
}

int load_extio_library(void)
{
char s[BUFFER_SIZE];
int errcod;
extio_handle=NULL;
extio_running=0;
if(ui.use_extio == 0)return 0;
get_library_filename(&s[0]);
if(ui.use_extio==0)return 1;
extio_handle=dlopen(s, RTLD_LAZY);
if( extio_handle==NULL) return 2;
OpenHW = (pOpenHW)dlsym(extio_handle, "OpenHW");
if(dlerror() != NULL)
  {
  errcod=3;
  OpenHW=NULL;
  goto errexit;
  }
extio_running=EXTIO_DLL_OPEN;
SetHWLO = (pSetHWLO) dlsym(extio_handle,"SetHWLO");
if(dlerror() != NULL)
  {
  errcod=4;
  goto errexit;
  }
CloseHW = (pCloseHW) dlsym(extio_handle,"CloseHW");
if(dlerror() != NULL)
  {
  errcod=5;
  goto errexit;
  }
SetCallback = (pSetCallback) dlsym(extio_handle, "SetCallback");
if(dlerror() != NULL)
  {
  errcod=6;
  goto errexit;
  }
InitHW = (pInitHW) dlsym(extio_handle, "InitHW");
if(dlerror() != NULL)
  {
  errcod=7;
  goto errexit;
  }
StopHW = (pStopHW) dlsym(extio_handle, "StopHW");
if(dlerror() != NULL)
  {
  errcod=8;
  goto errexit;
  }
ShowGUI = (pShowGUI) dlsym(extio_handle, "ShowGUI");
if(dlerror() != NULL)ShowGUI=NULL;
HideGUI = (pHideGUI) dlsym(extio_handle, "HideGUI");
if(dlerror() != NULL)HideGUI=NULL;
StartHW = (pStartHW) dlsym(extio_handle, "StartHW");
if(dlerror() != NULL)StartHW=NULL;
TuneChanged = (pTuneChanged) dlsym(extio_handle, "TuneChanged");
if(dlerror() != NULL)TuneChanged=NULL;
GetHWLO = (pGetHWLO) dlsym(extio_handle, "GetHWLO");
if(dlerror() != NULL)GetHWLO=NULL;
GetHWSR = (pGetHWSR) dlsym(extio_handle, "GetHWSR");
if(dlerror() != NULL)GetHWSR=NULL;
SetRFGain = (pSetRFGain) dlsym(extio_handle, "SetRFGain");
if(dlerror() != NULL)SetRFGain=NULL;
GetRFGain = (pGetRFGain) dlsym(extio_handle, "GetRFGain");
if(dlerror() != NULL)GetRFGain=NULL;
errcod=InitHW(extio_name, extio_model, &extio_type);
if(extio_type == 2)extio_type=4;
if(!errcod)
  {
  errcod=9;
  goto errexit;
  }
errcod=OpenHW();
if(!errcod)
  {
  errcod=10;
  goto errexit;
  }
return 0;
errexit:;
return errcod;
}

void unload_extio_library(void)
{
if(extio_type != 4 || (ui.use_extio&EXTIO_CALLBACK_ENABLE) != 0) 
  {
  if( (extio_running&EXTIO_DLL_OPEN) != 0)
    {
    CloseHW();
    extio_running&=~EXTIO_DLL_OPEN;
    }
  }
if(extio_running!=0)lirerr(539621);
if(extio_handle != NULL)
  {
  dlclose(extio_handle);
  }
}  



#endif

#if(OSNUM == OSNUM_WINDOWS)
#include "wdef.h"

void init_extio(void)
{
char s[BUFFER_SIZE];
char dirnam[BUFFER_SIZE];
WIN32_FIND_DATA file_data;
UINT charnum;
HANDLE find_handle;
char extio_lib_names[MAX_EXTIO_FILES][BUFFER_SIZE];
int i, num_extio;
int lineno;
i=0;
lineno=0;
lir_text(10,1,"Seaching libraries for ExtIO files");
lir_refresh_screen();
charnum=0;
while(i < 2)
  {
  if(i == 0) 
    {
    charnum=GetCurrentDirectory(BUFFER_SIZE, dirnam);
    }
  if(i == 1)
    {
    charnum=GetSystemDirectory(dirnam,BUFFER_SIZE);
    }
  sprintf(&dirnam[charnum],"\\Extio*.dll");
  find_handle=FindFirstFile(dirnam, &file_data);
  if(find_handle == INVALID_HANDLE_VALUE)goto nomore;
more:;
  if(i==0)
    {
    sprintf(extio_lib_names[lineno],"%s",file_data.cFileName);
    sprintf(s,"%2d %s",lineno,extio_lib_names[lineno]);
    lir_text(3,lineno+4,s);
    lineno++;
    }
  sprintf(extio_lib_names[lineno],"%s", dirnam); 
  sprintf(&extio_lib_names[lineno][charnum+1],"%s",file_data.cFileName);
  sprintf(s,"%2d %s",lineno,extio_lib_names[lineno]);
  lir_text(3,lineno+4,s);
  lineno++;
  if(FindNextFile(find_handle, &file_data))goto more;
nomore:;
  if(find_handle != INVALID_HANDLE_VALUE)FindClose(find_handle);
  i++;  
  }
if(lineno == 0)
  {
  lir_text(8,lineno+6,"No ExtIOxxx.dll file found");
  lir_text(5,lineno+8,press_any_key);
  await_keyboard();
  return;
  }  
else
  {
  lir_text(8,lineno+6,"Select ExtIO_xxx.dll file by line number:");
  num_extio=lir_get_integer(51,lineno+6,2,0,lineno-1);  
  if(kill_all_flag)return;
  }
write_extio_parfile(extio_lib_names[num_extio]);
}

int load_extio_library(void)
{
char s[BUFFER_SIZE];
int errcod;
extio_handle=NULL;
extio_running=0;
if(ui.use_extio == 0)return 0;
get_library_filename(&s[0]);
if(ui.use_extio==0)return 1;
extio_handle=LoadLibrary(s);
if(extio_handle == NULL)return 2;
OpenHW = (pOpenHW)GetProcAddress(extio_handle, "OpenHW");
if(!OpenHW)
  {
  errcod=3;
  goto errexit;
  }
extio_running=EXTIO_DLL_OPEN;
SetHWLO = (pSetHWLO) GetProcAddress(extio_handle, "SetHWLO");
if(!SetHWLO)
  {
  errcod=4;
  goto errexit;
  }
CloseHW = (pCloseHW) GetProcAddress(extio_handle, "CloseHW");
if(!CloseHW)
  {
  errcod=5;
  goto errexit;
  }
SetCallback = (pSetCallback) GetProcAddress(extio_handle, "SetCallback");
if(!SetCallback)
  {
  errcod=6;
  goto errexit;
  }
InitHW = (pInitHW) GetProcAddress(extio_handle, "InitHW");
if(!InitHW)
  {
  errcod=7;
  goto errexit;
  }
StopHW = (pStopHW) GetProcAddress(extio_handle, "StopHW");
if(!StopHW)
  {
  errcod=8;
  goto errexit;
  }
ShowGUI = (pShowGUI) GetProcAddress(extio_handle, "ShowGUI");
HideGUI = (pHideGUI) GetProcAddress(extio_handle, "HideGUI");
StartHW = (pStartHW) GetProcAddress(extio_handle, "StartHW");
TuneChanged = (pTuneChanged) GetProcAddress(extio_handle, "TuneChanged");
GetHWLO = (pGetHWLO) GetProcAddress(extio_handle, "GetHWLO");
SetRFGain = (pSetRFGain) GetProcAddress(extio_handle, "SetRFGain");
GetRFGain = (pGetRFGain) GetProcAddress(extio_handle, "GetRFGain");
GetHWSR = (pGetHWSR) GetProcAddress(extio_handle, "GetHWSR");
CoInitialize(NULL);
errcod=InitHW(extio_name, extio_model, &extio_type);
if(extio_type == 2)extio_type=4;
if(!errcod)
  {
  errcod=9;
  goto errexit;
  }
errcod=OpenHW();
if(!errcod)
  {
  errcod=10;
  goto errexit;
  }
return 0;
errexit:;
return errcod;
}

void unload_extio_library(void)
{
if(extio_type != 4 || (ui.use_extio&EXTIO_CALLBACK_ENABLE) != 0) 
  {
  if( (extio_running&EXTIO_DLL_OPEN) != 0)
    {
    CloseHW();
    extio_running&=~EXTIO_DLL_OPEN;
    }
  }  
if(extio_running!=0)lirerr(539621);
if(extio_handle != NULL)
  {
  FreeLibrary(extio_handle);
  }
}  



#endif

void confirm_extio(void)
{
char s[BUFFER_SIZE];
clear_screen();
lir_text(3,3,"This dynamic library is specified for RX input:");
get_library_filename(&s[0]);
lir_text(3,4,s);
lir_text(3,6,"Do you want to de-select this ExtIO input? (Y/N)");
gtkey:;
await_processed_keyboard();
if(lir_inkey == 'N' || kill_all_flag)return;
if(lir_inkey != 'Y')goto gtkey;
ui.use_extio=0;
ui.rx_addev_no=UNDEFINED_DEVICE_CODE;
}

void first_check_extio(void)
{
char s[BUFFER_SIZE], ss[80];
int line;
if(extio_error != 0)
  {
  get_library_filename(&s[0]);
  lir_text(3,3,"ERROR: dynamic library for ExtIO failed.");
  lir_text(3,5,s);
  lir_text(3,7,"Use of this library is disabled. Press W on the main menu");
  lir_text(3,8,"if you want to disable it permanently.");
  lir_text(8,11,press_any_key);
  await_processed_keyboard();
  ui.use_extio=0;
  return;
  }
if( (ui.use_extio&EXTIO_SELECTION_VERIFIED) == 0)
  {
  get_library_filename(&s[0]);
  lir_text(3,3,"First time open library.");
  lir_text(28,3,s);
  line=10;  
  sprintf(s,"Name:  %s",extio_name);
  lir_text(0,6,s);
  sprintf(s,"Model: %s",extio_model);
  lir_text(0,7,s);
  sprintf(s,"Type:");
  lir_text(0,8,s);
  switch(extio_type)
    {
    case 1:
    case 3:
    lir_text(7,8,"int 16bit dll callback");
    ui.rx_input_mode=IQ_DATA;
    goto prt_callb;
  
    case 4:
    lir_text(7,8,"Soundcard input");
    lir_text(2,line,
                 "Select the appropriate sound device with 'U' on the main menu");
    break;
  
    case 5:
    case 8:
    lir_text(7,8,"int 24bit dll callback");
    ui.rx_input_mode=IQ_DATA+BYTE_INPUT+DWORD_INPUT;
    goto prt_callb;
 
    case 6:
    lir_text(7,8,"int 32bit dll callback");
    ui.rx_input_mode=IQ_DATA+DWORD_INPUT+QWORD_INPUT;
    goto prt_callb;
  
    case 7:
    lir_text(7,8,"float 32bit dll callback");
    ui.rx_input_mode=IQ_DATA+DWORD_INPUT+FLOAT_INPUT;
prt_callb:;
    lir_text(2,line,
      "Set sampling speed and other parameters in the ExtIO window");
    ui.rx_addev_no=EXTIO_DEVICE_CODE;
    break;
  
    default:
    lirerr(1341);
    return; 
    }
  line+=2;
  if( (ui.use_extio&EXTIO_INVERTED) != 0)
    {
    s[0]=0;
    }
  else
    {
    sprintf(s,"non-");
    }
  sprintf(ss,"Spectrum: %sinverted",s);
  lir_text(2,line,ss);
  line+=2;
  lir_text(2,line,
      "Do not forget to press W on the main menu to make Linrad remember.");
  line+=2;      
  lir_text(7,line,press_any_key);
  await_processed_keyboard();  
  ui.use_extio|=EXTIO_SELECTION_VERIFIED;
  uiparm_change_flag=TRUE;
  }
}

void get_extio_speed(void)
{
if(extio_type != 4)
  {
  if(!GetHWSR)
    {
    lirerr(1316);
    return;
    }
  ui.rx_ad_speed=GetHWSR();
  extio_speed_changed=FALSE;
  switch(extio_type)
    {
    case 1:
    case 3:
// 16 bit integer    
    ui.rx_input_mode=IQ_DATA;
    snd[RXAD].framesize=4;
    break;
    
    case 5:
    case 8:
// 24 bit integer    
    ui.rx_input_mode=IQ_DATA+BYTE_INPUT+DWORD_INPUT;
    snd[RXAD].framesize=8;
    break;
    
    case 6:
// 32 bit integer    
    ui.rx_input_mode=IQ_DATA+DWORD_INPUT+QWORD_INPUT;
    snd[RXAD].framesize=8;
    break;
    
    case 7:
// 32 bit float
    ui.rx_input_mode=IQ_DATA+DWORD_INPUT+FLOAT_INPUT;
    snd[RXAD].framesize=8;
    break;
    
    default:
    lirerr(984546);
    break;
    }
  }    
ui.rx_rf_channels=1;
ui.rx_ad_channels=2;
}

void start_extio(void)
{
int siz;
long TuneFreq;
if( (extio_running&EXTIO_DLL_OPEN) == 0)
  {
  lirerr(522156);
  return;
  }
if(extio_running==EXTIO_DLL_OPEN)
  {
  if(extio_type != 4 || (ui.use_extio&EXTIO_CALLBACK_ENABLE) != 0) 
    {
    lir_init_event(EVENT_EXTIO_RXREADY);
    SetCallback(Callback_Winrad);
    }
  fg.gain_increment = 1;
  if(abs(fg.passband_direction) != 1)fg.passband_direction=1; 
  if(fg.passband_increment < 0.0001 || fg.passband_increment > 1.5)
    {
    fg.passband_increment=.01;
    }
  extio_running|=EXTIO_CALLBACK_RUNNING;
  }
TuneFreq = (long)(1000000L*fg.passband_center);
siz=StartHW(TuneFreq);
if(extio_type < 1 || extio_type > 8)
  {
  lirerr(1239);
  return;
  }
if(extio_type != 4)
  {
  if(!GetHWSR)
    {
    lirerr(1316);
    return;
    }
  if(ui.rx_ad_speed!=GetHWSR())extio_speed_changed=TRUE;
  timf1p_sdr=timf1p_pa;  
  snd[RXAD].block_frames=siz;
  snd[RXAD].block_bytes=snd[RXAD].framesize*siz;
  snd[RXAD].interrupt_rate=ui.rx_ad_speed/siz;
  }    
}
  
void stop_extio(void)
{
if(extio_type != 4 || (ui.use_extio&EXTIO_CALLBACK_ENABLE) != 0) 
  {
  if( (extio_running&EXTIO_CALLBACK_RUNNING) != 0)
    {
    StopHW();
    lir_close_event(EVENT_EXTIO_RXREADY);
    extio_running&=~EXTIO_CALLBACK_RUNNING;
    }
  }
}      

void update_extio_rx_freq(void)
{
long TuneFreq;
TuneFreq = (long)(1000000L*fg.passband_center);
SetHWLO(TuneFreq);
}

void update_extio_rx_gain(void)
{
if(fg.gain > RF_GAIN_UP_LIMIT) fg.gain = RF_GAIN_UP_LIMIT;
if(fg.gain < RF_GAIN_LOW_LIMIT) fg.gain = RF_GAIN_LOW_LIMIT;	
if(SetRFGain)SetRFGain(fg.gain);
}

void extio_input(void)
{
#if RUSAGE_OLD == TRUE
int local_workload_counter;
#endif
int errcod;
int rxin_local_workload_reset;
double dt1, read_start_time, total_reads;
int timing_loop_counter,timing_loop_counter_max,initial_skip_flag;
#if OSNUM == OSNUM_LINUX
clear_thread_times(THREAD_EXTIO_INPUT);
#endif
#if RUSAGE_OLD == TRUE
local_workload_counter=workload_counter;
#endif
errcod=0;
dt1=current_time();
fft1_block_timing();
fft1_hz_per_point=(float)ui.rx_ad_speed/fft1_size;
rxin_local_workload_reset=workload_reset_flag;
timing_loop_counter_max=snd[RXAD].interrupt_rate;
timing_loop_counter=2.5*timing_loop_counter_max;
screen_loop_counter_max=0.1*snd[RXAD].interrupt_rate;
if(screen_loop_counter_max==0)screen_loop_counter_max=1;
screen_loop_counter=screen_loop_counter_max;
read_start_time=current_time();
total_reads=0;
initial_skip_flag=1;
if(thread_command_flag[THREAD_SCREEN]!=THRFLAG_NOT_ACTIVE)
  {
  while(thread_status_flag[THREAD_SCREEN]!=THRFLAG_ACTIVE &&
        thread_status_flag[THREAD_SCREEN]!=THRFLAG_IDLE &&
        thread_status_flag[THREAD_SCREEN]!=THRFLAG_SEM_WAIT)
    {
    if(thread_command_flag[THREAD_EXTIO_INPUT] == THRFLAG_KILL)goto extio_exit;
    lir_sleep(10000);
    }
  }
thread_status_flag[THREAD_EXTIO_INPUT]=THRFLAG_ACTIVE;
while(thread_command_flag[THREAD_EXTIO_INPUT] == THRFLAG_ACTIVE)
  {
  if(rxin_local_workload_reset != workload_reset_flag)
    {
    if(timing_loop_counter <= timing_loop_counter_max)
      {
      rxin_local_workload_reset=workload_reset_flag;
      timing_loop_counter=5*timing_loop_counter_max;
      initial_skip_flag=1;
      }
    }
#if RUSAGE_OLD == TRUE
  if(local_workload_counter != workload_counter)
    {
    local_workload_counter=workload_counter;
    make_thread_times(THREAD_EXTIO_INPUT);
    }
#endif
  timing_loop_counter--;
  if(timing_loop_counter == 0)
    {
    timing_loop_counter=timing_loop_counter_max;
    if(initial_skip_flag != 0)
      {
      read_start_time=current_time();
      total_reads=0;
      initial_skip_flag=0;
      }
    else
      {
      dt1=current_time()-read_start_time;
      measured_ad_speed=total_reads*snd[RXAD].block_frames/dt1;
      }
    }
  while(thread_command_flag[THREAD_EXTIO_INPUT] == THRFLAG_ACTIVE &&
                                 !kill_all_flag && timf1p_sdr == timf1p_pa)
    {
    lir_await_event(EVENT_EXTIO_RXREADY);
    }
  while (!kill_all_flag && timf1p_sdr != timf1p_pa)
    {
    total_reads++;
    finish_rx_read();
    }                       
  if(kill_all_flag)goto extio_exit;
  if(extio_speed_changed)
    {
    while(keyboard_buffer_ptr != keyboard_buffer_used)lir_sleep(1000);
    keyboard_buffer[keyboard_buffer_used]='X';
    keyboard_buffer_ptr=(keyboard_buffer_ptr+1)&(KEYBOARD_BUFFER_SIZE-1);
    lir_sleep(10000);
    lir_set_event(EVENT_KEYBOARD);
    lir_sleep(100000);
    if(lir_inkey == 'X')extio_speed_changed=FALSE;
    if(kill_all_flag)goto extio_exit;
    }    
  }
extio_exit:;
if(errcod != 0)lirerr(errcod);
thread_status_flag[THREAD_EXTIO_INPUT]=THRFLAG_RETURNED;
while(thread_command_flag[THREAD_EXTIO_INPUT] != THRFLAG_NOT_ACTIVE)
  {
  lir_sleep(1000);
  }
}

